"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeGraphQLHandler = void 0;
const lru_1 = require("@graphile/lru");
const crypto_1 = require("crypto");
const grafast_1 = require("grafast");
const graphql_1 = require("graphql");
let lastString;
let lastHash;
const calculateQueryHash = (queryString) => {
    if (queryString !== lastString) {
        lastString = queryString;
        lastHash = (0, crypto_1.createHash)("sha1").update(queryString).digest("base64");
    }
    return lastHash;
};
function makeParseAndValidateFunction(schema) {
    const parseAndValidationCache = new lru_1.LRU({
        maxLength: 500,
    });
    let lastParseAndValidateQuery;
    let lastParseAndValidateResult;
    function parseAndValidate(query) {
        if (lastParseAndValidateQuery === query) {
            return lastParseAndValidateResult;
        }
        const hash = query.length > 500 ? calculateQueryHash(query) : query;
        const cached = parseAndValidationCache.get(hash);
        if (cached) {
            lastParseAndValidateQuery = query;
            lastParseAndValidateResult = cached;
            return cached;
        }
        const source = new graphql_1.Source(query, "GraphQL HTTP Request");
        let document;
        try {
            document = (0, graphql_1.parse)(source);
        }
        catch (e) {
            const result = {
                errors: [
                    new graphql_1.GraphQLError(e.message, null, undefined, undefined, undefined, e, undefined),
                ],
            };
            parseAndValidationCache.set(hash, result);
            lastParseAndValidateQuery = query;
            lastParseAndValidateResult = result;
            return result;
        }
        const errors = (0, graphql_1.validate)(schema, document);
        const result = errors.length
            ? { errors }
            : { document };
        parseAndValidationCache.set(hash, result);
        lastParseAndValidateQuery = query;
        lastParseAndValidateResult = result;
        return result;
    }
    return parseAndValidate;
}
const makeGraphQLHandler = (params) => {
    const { schema } = params;
    const parseAndValidate = makeParseAndValidateFunction(schema);
    const asString = true;
    return async (resolvedPreset, ctx, body) => {
        // Parse the body
        if (typeof body !== "object" || body == null) {
            throw new Error("Invalid body in request");
        }
        const { query, variables: variableValues, operationName, } = body;
        if (typeof query !== "string") {
            throw new Error("Invalid query");
        }
        if (typeof operationName !== "string" && operationName != null) {
            throw new Error("Invalid operationName");
        }
        if (variableValues != null && typeof variableValues !== "object") {
            throw new Error("Invalid variables");
        }
        const { errors, document } = parseAndValidate(query);
        if (errors) {
            return { type: "graphql", statusCode: 200, payload: { errors } };
        }
        const args = {
            schema,
            document,
            rootValue: null,
            contextValue: Object.create(null),
            variableValues,
            operationName,
        };
        await (0, grafast_1.hookArgs)(args, ctx, resolvedPreset);
        try {
            const result = await (0, grafast_1.execute)(args, resolvedPreset);
            if ((0, grafast_1.isAsyncIterable)(result)) {
                return {
                    type: "graphqlIncremental",
                    statusCode: 200,
                    iterator: result,
                    asString,
                };
            }
            return { type: "graphql", statusCode: 200, payload: result, asString };
        }
        catch (e) {
            console.error(e);
            return {
                type: "graphql",
                statusCode: 500,
                payload: {
                    errors: [new graphql_1.GraphQLError(e.message)],
                    extensions: args.rootValue?.[grafast_1.$$extensions],
                },
            };
        }
    };
};
exports.makeGraphQLHandler = makeGraphQLHandler;
//# sourceMappingURL=graphql.js.map