"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.grafserv = void 0;
const tslib_1 = require("tslib");
const eventemitter3_1 = tslib_1.__importDefault(require("eventemitter3"));
const grafast_1 = require("grafast");
const graphile_config_1 = require("graphile-config");
const graphql_1 = require("graphql");
const graphiql_js_1 = require("./graphiql.js");
const graphql_js_1 = require("./graphql.js");
function getBodyFromRequest(req, maxLength) {
    return new Promise((resolve, reject) => {
        req.setEncoding("utf8");
        let data = "";
        const handleData = (chunk) => {
            data += chunk;
            if (data.length > maxLength) {
                req.off("end", done);
                req.off("error", reject);
                req.off("data", handleData);
                // TODO: validate this approach
                reject(new Error("Too much data"));
            }
        };
        const done = () => {
            resolve(data);
        };
        req.on("end", done);
        req.on("error", reject);
        req.on("data", handleData);
    });
}
function handleErrors(payload) {
    if ("errors" in payload && payload.errors) {
        payload.errors = payload.errors.map((e) => {
            const obj = e instanceof graphql_1.GraphQLError
                ? e.toJSON()
                : { message: e.message, ...e };
            return Object.assign(obj, {
                message: (0, grafast_1.stripAnsi)(obj.message),
                extensions: { stack: (0, grafast_1.stripAnsi)(e.stack ?? "").split("\n") },
            });
        });
    }
}
function optionsFromConfig(config) {
    const { graphqlPath = "/graphql", graphiql = true, graphiqlOnGraphQLGET = true, graphiqlPath = "/", watch = false, eventStreamRoute = "/graphql/stream", maxRequestLength = 100_000, } = config.server ?? {};
    return {
        graphqlPath,
        graphiql,
        graphiqlOnGraphQLGET,
        graphiqlPath,
        watch,
        eventStreamRoute,
        maxRequestLength,
    };
}
function grafserv(preset, initialParams) {
    /**
     * Mutable options, change them by calling `setParams`. Don't dereference
     * properties as they should be dynamic.
     */
    const resolvedPreset = (0, graphile_config_1.resolvePresets)([preset]);
    let dynamicOptions = optionsFromConfig(resolvedPreset);
    const sendResult = (res, handlerResult) => {
        switch (handlerResult.type) {
            case "graphql": {
                const { payload, statusCode = 200, asString } = handlerResult;
                handleErrors(payload);
                res.writeHead(statusCode, {
                    "Content-Type": "application/json",
                    ...(dynamicOptions.watch
                        ? {
                            "X-GraphQL-Event-Stream": dynamicOptions.eventStreamRoute,
                        }
                        : null),
                });
                const payloadString = (0, grafast_1.stringifyPayload)(payload, asString);
                res.end(payloadString);
                break;
            }
            case "graphqlIncremental": {
                const { iterator, statusCode = 200, asString } = handlerResult;
                res.writeHead(statusCode, {
                    "Content-Type": 'multipart/mixed; boundary="-"',
                    "Transfer-Encoding": "chunked",
                    ...(dynamicOptions.watch
                        ? {
                            "X-GraphQL-Event-Stream": dynamicOptions.eventStreamRoute,
                        }
                        : null),
                });
                (async () => {
                    // TODO: use manual looping so the iterable can be aborted without awaiting the promise
                    try {
                        for await (const payload of iterator) {
                            handleErrors(payload);
                            const payloadString = (0, grafast_1.stringifyPayload)(payload, asString);
                            res.write("\r\n---\r\nContent-Type: application/json\r\n\r\n" +
                                payloadString);
                        }
                    }
                    finally {
                        res.write("\r\n-----\r\n");
                        res.end();
                    }
                })().catch((e) => {
                    console.error(`Error occurred when streaming result: ${e}`);
                    try {
                        res.end();
                    }
                    catch (e2) {
                        console.error(`A further error occurred when terminating the request: ${e2}`);
                    }
                });
                break;
            }
            case "text":
            case "html": {
                const { payload, statusCode = 200 } = handlerResult;
                res.writeHead(statusCode, {
                    "Content-Type": handlerResult.type === "html"
                        ? "text/html; charset=utf-8"
                        : "text/plain; charset=utf-8",
                });
                res.end(payload);
                break;
            }
            case "event-stream": {
                const { payload: stream, statusCode = 200 } = handlerResult;
                // Making sure these options are set.
                res.req.socket.setTimeout(0);
                res.req.socket.setNoDelay(true);
                res.req.socket.setKeepAlive(true);
                // Set headers for Server-Sent Events.
                // Don't buffer EventStream in nginx
                res.setHeader("X-Accel-Buffering", "no");
                res.setHeader("Content-Type", "text/event-stream");
                res.setHeader("Cache-Control", "no-cache, no-transform");
                if (res.req.httpVersionMajor >= 2) {
                    // NOOP
                }
                else {
                    res.setHeader("Connection", "keep-alive");
                }
                res.writeHead(statusCode);
                // Creates a stream for the response
                let stopped = false;
                const send = (event) => {
                    if (stopped) {
                        return;
                    }
                    let payload = "";
                    if (event.event) {
                        payload += `event: ${event.event}\n`;
                    }
                    if (event.id) {
                        payload += `id: ${event.id}\n`;
                    }
                    if (event.retry) {
                        payload += `retry: ${event.retry}\n`;
                    }
                    if (event.data != null) {
                        payload += `data: ${event.data.replace(/\n/g, "\ndata: ")}\n`;
                    }
                    payload += "\n";
                    res.write(payload);
                    // Technically we should see if `.write()` returned false, and if so we
                    // should pause the stream. However, since our stream is coming from
                    // watch mode, we find it unlikely that a significant amount of data
                    // will be buffered (and we don't recommend watch mode in production),
                    // so it doesn't feel like we need this currently. If it turns out you
                    // need this, a PR would be welcome.
                    if (typeof res.flush === "function") {
                        // https://github.com/expressjs/compression#server-sent-events
                        res.flush();
                    }
                    else if (typeof res.flushHeaders === "function") {
                        res.flushHeaders();
                    }
                };
                // Notify client that connection is open.
                send({ event: "open" });
                // Process stream
                const iterator = stream[Symbol.asyncIterator]();
                const waitNext = () => {
                    const n = iterator.next();
                    n.then((r) => {
                        if (!r.done || r.value) {
                            send(r.value);
                        }
                        if (r.done) {
                            res.end();
                            cleanup();
                            return;
                        }
                        else {
                            if (!stopped) {
                                waitNext();
                            }
                        }
                    }, (error) => {
                        console.error("Error occurred processing event stream:", error);
                        send({
                            event: "error",
                            data: JSON.stringify({
                                errors: [
                                    { message: "Error occurred processing event stream" },
                                ],
                            }),
                        });
                        res.end();
                        cleanup();
                    });
                };
                waitNext();
                // Clean up when connection closes.
                const cleanup = () => {
                    stopped = true;
                    try {
                        iterator.return?.();
                    }
                    catch {
                        /* nom nom nom */
                    }
                    res.req.removeListener("close", cleanup);
                    res.req.removeListener("finish", cleanup);
                    res.req.removeListener("error", cleanup);
                };
                res.req.on("close", cleanup);
                res.req.on("finish", cleanup);
                res.req.on("error", cleanup);
                break;
            }
            default: {
                const never = handlerResult;
                console.error(`Did not understand '${never}' passed to sendResult`);
                res.writeHead(500, { "Content-Type": "text/plain; charset=utf-8" });
                res.end("Unexpected input to sendResult");
            }
        }
    };
    const makeErrorHandler = (req, res, next) => {
        if (typeof next === "function") {
            return next;
        }
        return (e) => {
            console.error(e);
            sendResult(res, {
                statusCode: 500,
                type: "text",
                payload: "Internal server error",
            });
        };
    };
    const eventEmitter = new eventemitter3_1.default();
    function makeStream() {
        const queue = [];
        let finished = false;
        const bump = () => {
            const next = queue.shift();
            if (next) {
                next.resolve({
                    done: false,
                    value: { event: "change", data: "schema" },
                });
            }
        };
        const flushQueue = (e) => {
            const entries = queue.splice(0, queue.length);
            for (const entry of entries) {
                if (e) {
                    entry.reject(e);
                }
                else {
                    entry.resolve({ done: true });
                }
            }
        };
        eventEmitter.on("newSchema", bump);
        return {
            [Symbol.asyncIterator]() {
                return this;
            },
            next() {
                if (finished) {
                    return Promise.resolve({
                        done: true,
                    });
                }
                return new Promise((resolve, reject) => {
                    queue.push({ resolve, reject });
                });
            },
            return() {
                finished = true;
                if (queue.length) {
                    flushQueue();
                }
                return Promise.resolve({
                    done: true,
                });
            },
            throw(e) {
                if (queue.length) {
                    flushQueue(e);
                }
                return Promise.reject(e);
            },
        };
    }
    function addHandlers(r) {
        eventEmitter.emit("newSchema", r.schema);
        return {
            ...r,
            graphqlHandler: (0, graphql_js_1.makeGraphQLHandler)(r),
            graphiqlHandler: (0, graphiql_js_1.makeGraphiQLHandler)(r),
        };
    }
    let serverParams = initialParams == null
        ? (0, grafast_1.defer)()
        : (0, grafast_1.isPromiseLike)(initialParams)
            ? initialParams.then(addHandlers)
            : addHandlers(initialParams);
    const middleware = (req, res, next) => {
        const handleError = makeErrorHandler(req, res, next);
        // TODO: consider allowing GraphQL queries over 'GET'
        if (req.url === dynamicOptions.graphqlPath && req.method === "POST") {
            (async () => {
                const bodyRaw = await getBodyFromRequest(req, dynamicOptions.maxRequestLength);
                // TODO: this parsing is unsafe (it doesn't even check the
                // content-type!) - replace it with V4's behaviour
                const body = JSON.parse(bodyRaw);
                const sP = (0, grafast_1.isPromiseLike)(serverParams)
                    ? await serverParams
                    : serverParams;
                const result = await sP.graphqlHandler(resolvedPreset, {
                    httpRequest: req,
                }, body);
                sendResult(res, result);
            })().catch((e) => {
                // Special error handling for GraphQL route
                try {
                    console.error("An error occurred whilst attempting to handle the GraphQL request:");
                    console.dir(e);
                    sendResult(res, {
                        type: "graphql",
                        payload: { errors: [e] },
                        statusCode: 500,
                    });
                }
                catch (e2) {
                    console.error("An error occurred whilst telling the user than an error occurred:", e2);
                }
            });
            return;
        }
        // TODO: handle 'HEAD' requests
        if (dynamicOptions.graphiql &&
            (req.url === dynamicOptions.graphiqlPath ||
                (dynamicOptions.graphiqlOnGraphQLGET &&
                    req.url === dynamicOptions.graphqlPath)) &&
            req.method === "GET") {
            (async () => {
                const sP = (0, grafast_1.isPromiseLike)(serverParams)
                    ? await serverParams
                    : serverParams;
                const result = await sP.graphiqlHandler();
                sendResult(res, result);
            })().catch(handleError);
            return;
        }
        if (dynamicOptions.watch &&
            req.url === dynamicOptions.eventStreamRoute &&
            req.method === "GET") {
            (async () => {
                const stream = makeStream();
                sendResult(res, {
                    type: "event-stream",
                    payload: stream,
                    statusCode: 200,
                });
            })().catch(handleError);
            return;
        }
        // Not handled
        if (next) {
            return next();
        }
        else {
            console.log(`Unhandled ${req.method} to ${req.url}`);
            sendResult(res, {
                type: "text",
                payload: `Could not process ${req.method} request to ${req.url} ─ please POST requests to ${dynamicOptions.graphqlPath}`,
                statusCode: 404,
            });
            return;
        }
    };
    const releaseHandlers = [];
    return {
        handler: middleware,
        async release() {
            for (const handler of releaseHandlers) {
                try {
                    await handler();
                }
                catch (e) {
                    /* nom nom nom */
                }
            }
        },
        onRelease(cb) {
            releaseHandlers.push(cb);
        },
        setParams(newParams) {
            const newParamsWithHandlers = addHandlers(newParams);
            if (
            // If serverParams was deferred, resolve it
            serverParams !== null &&
                "resolve" in serverParams &&
                typeof serverParams.resolve === "function") {
                serverParams.resolve(newParamsWithHandlers);
            }
            dynamicOptions = optionsFromConfig(newParams.resolvedPreset);
            serverParams = newParamsWithHandlers;
        },
    };
}
exports.grafserv = grafserv;
//# sourceMappingURL=index.js.map